do.gene.set.analysis <-
    function(EventsBySample,
             Scores,
             GeneSizes,
             GeneSets,
             ##passenger rates
             passenger.rates = t(data.frame(0.55*rep(1.0e-6,25))),
             ##coverage object
             Coverage,
             ID2name,
	     BH = TRUE,
             ##T/F whether to do gene method
             gene.method = TRUE,
             ##T/F whether to do subject method
             perm.null.method = TRUE,
             ##T/F whether to do heterogeneous subject method
             perm.null.het.method = TRUE,
             ##T/F whether to do passenger null method
             pass.null.method = TRUE,
             ##T/F whether to do heterogeneous passenger null method
             pass.null.het.method = TRUE)
{
    ##make sure the passenger.rates object has the proper names
    names(passenger.rates) <-
        c("C.in.CpG.to.G", "C.in.CpG.to.A", "C.in.CpG.to.T",
          "G.in.CpG.to.C", "G.in.CpG.to.A", "G.in.CpG.to.T",
          "G.in.GpA.to.C", "G.in.GpA.to.A", "G.in.GpA.to.T",
          "C.in.TpC.to.G", "C.in.TpC.to.A", "C.in.TpC.to.T",
          "A.to.C", "A.to.G", "A.to.T",
          "C.not.in.CpG.or.TpC.to.G", "C.not.in.CpG.or.TpC.to.A",
          "C.not.in.CpG.or.TpC.to.T",
          "G.not.in.CpG.or.GpA.to.C", "G.not.in.CpG.or.GpA.to.A",
          "G.not.in.CpG.or.GpA.to.T",
          "T.to.C", "T.to.G", "T.to.A",
          "ins.del")

    ##transform the GeneSets object into a list, if it is not already a list
    GeneSets <- as.list(GeneSets)
    
    ##get the unique samples that have at least one mutation
    GoodSamples <- unique(EventsBySample[,"Sample"])

    ##get the gene symbols
    GeneNames <- rownames(GeneSizes)
    GeneSymb <- shorten.gene.names(GeneNames)

    ##make sure that the Coverage and GeneSizes objects have the same order of rows
    Coverage <- Coverage[GeneNames, ]
    
    ##if there are gene IDs which are not in the gene names, then use the ID2name to change them around
    IDs <- unique(unlist(GeneSets))
    if(length(setdiff(IDs, GeneSymb)) > 0)
    {
        GeneSets <- sapply(GeneSets,
                           function(set, i2n) {i2n[set]},
                           ID2name)
    }
    ##now eliminate all the NAs
    GeneSets <- sapply(GeneSets, function(x) {x[!is.na(x)]})

    EventsBySampleBySet <-
        get.EventsBySampleBySet.from.EventsBySample(EventsBySample,
                                                    GeneSets,
                                                    GoodSamples)
    
    ##get number of events in each sample
    EventsBySampleMut <-
        EventsBySample[EventsBySample$Event == "Mut",]

    nr.events.sample <-
        sapply(split(EventsBySampleMut$Sample,
                     EventsBySampleMut$Sample),
               length)
    nr.events.sample <- nr.events.sample[GoodSamples]

    ##gene-oriented method
    if(gene.method)
    {
        print("Gene method!")
        print(date())

        GeneScores <- data.frame(matrix(0,length(unique(GeneSymb)),1))
        rownames(GeneScores) <- unique(GeneSymb)
        colnames(GeneScores) <- "logLRT"

        transcripts2genes <- shorten.gene.names(rownames(Scores))
        names(transcripts2genes) <- rownames(Scores)
        ##get only the first transcript if there are multiple transcripts per gene
        Scores.short <- rep(0, length(unique(transcripts2genes)))
        names(Scores.short) <- unique(transcripts2genes)
        for(g in names(Scores.short))
        {
            t <- names(transcripts2genes[transcripts2genes == g])
            Scores.short[g] <- Scores[t[1],"logLRT"]
        }

        ##print(head(Scores.short))
        ##print(head(GeneScores))

        ## put non-zero LRT scores in GeneScores data frame
        GeneScores[match(names(Scores.short),
                         rownames(GeneScores)),"logLRT"] <-
                             Scores.short

        ##get p-values using the GSEA approach
        GeneScores.logLRT <- GeneScores[,"logLRT", drop = FALSE]

        p.values.gene <- sapply(GeneSets,runGSEA,
                                data=GeneScores.logLRT,
                                datacol=1,
                                IDinRowNames=TRUE, absolute=TRUE,
                                type= "f", alternative="mixed",
                                ranks.only=TRUE, nsim=1)
    }
    else
    {
        p.values.gene <- NA
    }

    ##get p-values using the subject approach
    if(perm.null.method)
    {
        print("Permutation null w/o heterogeneity")
        print(date())

        p.values.subject <-
          perm.null.p.values(EventsBySampleBySet =
                             EventsBySampleBySet,
                             GoodSamples = GoodSamples,
                             GeneSymb = GeneSymb,
                             nr.events.sample =  nr.events.sample,
                             GeneSets = GeneSets)
    }
    else
    {
        p.values.subject <- NA
    }

    ##get p-values using the subject "heterogeneous" approach
    if(perm.null.het.method)
    {
        print("Permutation null w/ heterogeneity")
        print(date())

        ##dyn.load("get.cdf.het.berns.so")

        p.values.subject.het <-
            perm.null.het.p.values(EventsBySampleBySet =
                                   EventsBySampleBySet,
                                   GoodSamples = GoodSamples,
                                   nr.events.sample =  nr.events.sample,
                                   GeneSets = GeneSets,
                                   GeneSymb = GeneSymb)
    }
    else
    {
        p.values.subject.het <- NA
    }

    ##passenger null method
    if(pass.null.method)
    {
        print("Passenger null w/o heterogeneity")
        print(date())
        
        p.values.pass.null <-
            pass.null.p.values(GeneSets = GeneSets,
                               GoodSamples = GoodSamples,
                               GeneSymb = GeneSymb,
                               nr.events.sample = nr.events.sample,
                               passenger.rates = passenger.rates,
                               Coverage = Coverage,
                               GeneSizes = GeneSizes,
                               EventsBySample = EventsBySample,
                               EventsBySampleBySet = EventsBySampleBySet)
    }
    else
    {
        p.values.pass.null <- NA
    }


    ##heterogeneous passenger null method
    if(pass.null.het.method)
    {
        print("Passenger null w/ heterogeneity")
        print(date())

        ##dyn.load("get.cdf.het.berns.so")

        p.values.pass.null.het <-
            pass.null.het.p.values(GeneSets = GeneSets,
                                   GoodSamples = GoodSamples,
                                   GeneSymb = GeneSymb,
                                   nr.events.sample = nr.events.sample,
                                   passenger.rates = passenger.rates,
                                   Coverage = Coverage,
                                   GeneSizes = GeneSizes,
                                   EventsBySample = EventsBySample,
                                   EventsBySampleBySet = EventsBySampleBySet)
    }
    else
    {
        p.values.pass.null.het <- NA
    }

    print(date())

    results <- data.frame(p.values.gene = p.values.gene,
                          q.values.gene = p2q(p.values.gene, 
			  		      BH = BH),
                          p.values.perm.null = p.values.subject,
                          q.values.perm.null = p2q(p.values.subject, 
			  		           BH = BH),
                          p.values.perm.null.het = p.values.subject.het,
                          q.values.perm.null.het = p2q(p.values.subject.het, 
			  			       BH = BH),
                          p.values.pass.null = p.values.pass.null,
                          q.values.pass.null = p2q(p.values.pass.null, 
			  		           BH = BH),
                          p.values.pass.null.het = p.values.pass.null.het,
                          q.values.pass.null.het = p2q(p.values.pass.null.het, 
			  			       BH = BH))

    results
}

