##return objects of the form EventsBySampleBrain etc (actually a list
##of such objects, of length nr.iterations),
##as well an "augmented" GeneSets object, with the fake spiked-in sets
##with their fake genes
##nr.iterations = nr of iterations
##GoodSamples = see above
##perc.samples = percentages of samples in which the gene-sets are spiked in
##spiked.set.sizes = sizes of sets spiked in
##mutations.only = flag on whether should just use mutational event
##(if FALSE, also use amplifications and deletions)
##pass.null = if TRUE, use passenger null instead of permutation null
sim.data <- function(GeneSets,
                     GeneSizes,
                     Coverage,
                     Mutations,
                     genes,
                     passenger.rates,
                     nr.iterations = 100,
                     EventsBySample,
                     EventsBySampleBySet,
                     GoodSamples,
                     pass.null = FALSE,
                     perc.samples = c(25, 50, 75, 90),
                     spiked.set.sizes = c(20, 100, 500),
                     KnownMountains = c("EGFR","SMAD4","KRAS",
                     "TP53","CDKN2A","MYC","MYCN","PTEN","RB1"),
                     exclude.mountains=TRUE)
{
    ##get the gene symbols
    GeneNames <- rownames(GeneSizes)
    GeneSymb <- shorten.gene.names(GeneNames)
    
    ##make sure that the Coverage and GeneSizes objects have the same order of rows
    Coverage <- Coverage[GeneNames, ]
    
    EventsBySample <- EventsBySample[EventsBySample$Event=="Mut",]
    
    ##remove mutations in mountains
    if (exclude.mountains) {
        EventsBySample <-
            EventsBySample[!(EventsBySample[, "Symbol"] %in% KnownMountains), ]
    }
    
    ##create vectors of names of spiked sets and names of spiked genes ("fake sets" and "fake genes"),
    fake.sets <- c()
    fake.genes <- c()
    ##also create vector of names of spiked transcripts (simply add "@@00" after gene name)
    fake.transcripts <- c()
    if(length(spiked.set.sizes) > 0)
    {
        for(size in spiked.set.sizes)
        {
            for(perc in perc.samples)
            {
                ##create name for spiked set
                fake.set <- paste("gene.set", size, perc, sep = ".")
                ##add to vector of names of spiked-in sets
                fake.sets <- c(fake.sets,
                               fake.set)
                ##create genes inside spiked set and add the set to GeneSets object
                fake.genes.set <- paste("gene", size, perc, "nr", 1:size, sep = ".")
                ##add the genes to the fake.genes object
                fake.genes <- c(fake.genes, fake.genes.set)
                GeneSets[[fake.set]] <- fake.genes.set
            }
        }
        fake.transcripts <- paste(fake.genes, "@@00", sep="")
    }
    
    ##get number of events in each sample
    nr.events.sample <- rep(0, length(GoodSamples))
    names(nr.events.sample) <- GoodSamples
    
    for(sample in GoodSamples)
    {
        nr.events.sample[sample] <-
            nrow(EventsBySample[EventsBySample[, "Sample"] == sample, ])
    }
    
    ##create objects for simulated data
    sim.GeneSizes <- list()
    sim.Coverage <- list()
    sim.EventsBySample <- list()
    ##sim.EventsBySampleBySet <- list()
    sim.Mutations <- list()
    sim.Scores <- list()
    
    ##create a vector of correspondence between transcript name and gene-name
    transcripts <- c(rownames(Coverage), fake.transcripts)
    transcripts2genes <- shorten.gene.names(transcripts)
    names(transcripts2genes) <- transcripts

    ##create a vector of correspondence between mutation types and contexts
    mutations2contexts <- c(rep(c("C.in.CpG","G.in.CpG","G.in.GpA","C.in.TpC",
                                  "A","C.not.in.CpG.or.TpC","G.not.in.CpG.or.GpA",
                                  "T"), each = 3), "ins.del")
    names(mutations2contexts) <- names(passenger.rates)
    
    ##create a vector of correspondence between mutation types preceded by
  ##"MutationsDiscovery." and mutation types
    mutations2add <- names(passenger.rates)
    
    names(mutations2add) <- paste("MutationsDiscovery.",
                                  names(passenger.rates),
                                  sep="")
    
  for(i in 1:nr.iterations)
  {
      print(paste("Currently simulating data: Iteration #", i, sep=""))
      print(date())
      
      sim.EventsBySample.i <- c()
      sim.Mutations.i <- data.frame(matrix(0, nrow = length(transcripts),
                                           ncol = ncol(Mutations)))
      rownames(sim.Mutations.i) <- transcripts
      colnames(sim.Mutations.i) <- colnames(Mutations)
      
      ##if there are no spiked-in genes, keep same Coverage and GeneSizes objects
      sim.GeneSizes.i <- GeneSizes
      sim.Coverage.i <- Coverage
      
      ##if there are spiked-in genes, then these objects will be augmented
      ##by some randomly selected coverages and sizes for the spike-ins
      if(length(spiked.set.sizes) > 0)
      {
          ##randomly sample some genes
          sampled.genes <- sample(rownames(GeneSizes),
                                  length(fake.transcripts))

          sim.GeneSizes.i[fake.transcripts,] <-
              GeneSizes[sampled.genes,]
          sim.Coverage.i[fake.transcripts,] <-
              Coverage[sampled.genes,]
        }
      
      sim.GeneSizes[[i]] <- sim.GeneSizes.i
      sim.Coverage[[i]] <- sim.Coverage.i
      
      ##create object with expected number of mutations, given the coverages and
      ##background rates
      sim.exp.nr.mut.i <-
          as.matrix(sim.Coverage.i[, c(rep(1:8, each = 3), 9)]) *
              matrix(rep(unlist(passenger.rates), nrow(sim.Coverage.i)),
                     ncol = 25, byrow = TRUE)
      rownames(sim.exp.nr.mut.i) <- rownames(sim.Coverage.i)
      colnames(sim.exp.nr.mut.i) <- names(passenger.rates)
      
      if(pass.null)
      {
          ##get sample-specific constants we're multiplying the passenger null by
          sample.constants <- matrix(rep(nr.events.sample/mean(nr.events.sample),
                                         each=length(passenger.rates)),ncol=length(GoodSamples))
          colnames(sample.constants) <- GoodSamples
          
          ##make matrix of sample-specific passenger rates
          ##each column represents a sample
          ##(so columns differ just by a multiplicative constant)
          passenger.rates.mat <- matrix(rep(as.numeric(passenger.rates),
                                            length(GoodSamples)),
                                        ncol=length(GoodSamples))
          colnames(passenger.rates.mat) <- GoodSamples
          rownames(passenger.rates.mat) <- names(passenger.rates)
          passenger.rates.mat <- passenger.rates.mat*sample.constants
          
          ##distribute events over genes for non-spiked-in gene-sets
          for(sample in names(nr.events.sample))
          {
              ##get mutations
              ##print(sample)
              Mut <- cma.simulator(gene.sizes = GeneSizes,
                                   passenger.rates =
                                   t(data.frame(passenger.rates.mat[,sample])),
                                   coverage = Coverage/length(GoodSamples),
                                   eliminate.noval = FALSE)
              
              genes.with.events <- c()
              contexts <- c()
              
              ##get the rownames with mutations
              Mut.nonzero <- Mut[rowSums(Mut[,1:25]) > 0,]
              
              ##get the number of events for each gene
              nr.events.for.gene <- rowSums(Mut.nonzero[,1:25])

              ##write out the genes that have events
              genes.with.events <- rep(rownames(Mut.nonzero),
                                       nr.events.for.gene)

              ##transform Mut.nonzero into a vector to make it easier to manipulate
              Mut.nonzero.as.vect <-
                  as.vector(t(as.matrix(Mut.nonzero[,1:25])))
              ##get the types of events for each gene
              types <- rep(rep(colnames(Mut.nonzero[,1:25]),
                               times = nrow(Mut.nonzero)),
                           Mut.nonzero.as.vect)

              ##get the number of events of each type
              events.per.type <-
                  Mut.nonzero.as.vect[Mut.nonzero.as.vect > 0]

              contexts <- types

              ##add stuff to sim.EventsBySample.i
              sim.EventsBySample.i <- rbind(sim.EventsBySample.i,
                                            cbind("Mut", sample,
                                                  shorten.gene.names(genes.with.events),
                                                  mutations2add[contexts]))

              ##add mutations to mutation object
              sim.Mutations.i[rownames(Mut),1:50] <-
                sim.Mutations.i[rownames(Mut),1:50] +
                  Mut[rownames(Mut),1:50]

              ##see if sample is altered for each of the spiked-in gene-sets
              ##considered
              unifs <- matrix(runif(length(perc.samples)*length(spiked.set.sizes), 0, 1),
                              ncol = length(perc.samples), nrow = length(spiked.set.sizes))

              ##make matrix of prob. that a gene-set is altered in this sample
              prob.set.alt <-
                  matrix(rep(perc.samples, length(spiked.set.sizes)),
                         ncol = length(perc.samples))/100

              ##get the entries where unifs < prob.set.alt -> these will correspond to the altered spiked-in sets
              alt.spiked <- which(prob.set.alt > unifs, arr.ind = TRUE)
              alt.spiked[,"col"] <- perc.samples[alt.spiked[,"col"]]
              alt.spiked[,"row"] <- spiked.set.sizes[alt.spiked[,"row"]]
              ##get the actual names of the spiked-in sets
              alt.spiked.sets <- sapply(as.data.frame(t(alt.spiked)),
                                        function(vect) {paste("gene.set.",
                                                              paste(vect,collapse="."),
                                                              sep="")})

              for(set in alt.spiked.sets)
              {
                  ##if sample is altered, randomly choose a gene
                  ##for it to be altered in - weigh by total expected number
                  ##of mutations

                  ##get genes in this spiked-in set
                  genes.in.spiked.set <-
                      GeneSets[[set]]

                  ##get corresponding transcripts
                  transc.in.spiked.set <-
                      paste(genes.in.spiked.set, "@@00", sep="")

                  spiked.in.exp.nr.mut <-
                      rowSums(sim.exp.nr.mut.i[transc.in.spiked.set, ])

                  altered.gene <-
                      sample(genes.in.spiked.set, 1,
                             prob = spiked.in.exp.nr.mut)

                  ##get altered transcript for altered gene
                  altered.gene <-
                      names(transcripts2genes[transcripts2genes == altered.gene])

                  genes.with.events <-
                      c(genes.with.events, altered.gene)

                  ##add stuff to sim.EventsBySample.i and sim.Mutations.i objects
                  ##use weights to pick a mutation for this alteration
                  ##the weights are equal to the expected number of mutations
                  ##under the null
                  mutation.gene <- sample(names(passenger.rates),1,replace=FALSE,
                                          prob=sim.exp.nr.mut.i[altered.gene, ])

                  ##add stuff to sim.EventsBySample.i
                  sim.EventsBySample.i <- rbind(sim.EventsBySample.i,
                                                c("Mut", sample,
                                                  shorten.gene.names(altered.gene),
                                                  mutation.gene))

                  ##add stuff to sim.Mutations.i
                  sim.Mutations.i[altered.gene,
                                  paste("MutationsDiscovery.",mutation.gene,sep="")] <-
                                      sim.Mutations.i[altered.gene,
                                                      paste("MutationsDiscovery.",
                                                            mutation.gene,sep="")] + 1

              }
          }
      }
      else
      {
          ##distribute events over genes for non-spiked-in gene-sets
          for(sample in names(nr.events.sample))
          {
              ##sample transcripts with events from all transcripts
              genes.with.events <-
                  sample(rownames(Coverage), nr.events.sample[sample], replace = FALSE)

              ##assign mutations to the genes
              ##mutation.genes <-
              ##    sample(EventsBySample$MutationClass[EventsBySample$Sample == sample])
              ##names(mutation.genes) <- genes.with.events

              ##see if sample is altered for each of the spiked-in gene-sets
              ##considered
              unifs <- matrix(runif(length(perc.samples)*length(spiked.set.sizes), 0, 1),
                              ncol = length(perc.samples), nrow = length(spiked.set.sizes))

              ##make matrix of prob. that a gene-set is altered in this sample
              prob.set.alt <-
                  matrix(rep(perc.samples, length(spiked.set.sizes)),
                         ncol = length(perc.samples))/100

              ##get the entries where unifs < prob.set.alt -> these will correspond to the altered spiked-in sets
              alt.spiked <- which(prob.set.alt > unifs, arr.ind = TRUE)
              alt.spiked[,"col"] <- perc.samples[alt.spiked[,"col"]]
              alt.spiked[,"row"] <- spiked.set.sizes[alt.spiked[,"row"]]
              ##get the actual names of the spiked-in sets
              alt.spiked.sets <- sapply(as.data.frame(t(alt.spiked)),
                                        function(vect) {paste("gene.set.",
                                                              paste(vect,collapse="."),
                                                              sep="")})
              ##randomly chose a gene to be altered from each spiked-in set
              altered.spiked.genes <- sapply(GeneSets[alt.spiked.sets], "sample", 1)
              ##now get transcripts for these genes
              genes.with.events.spiked <- paste(altered.spiked.genes, "@@00", sep="")

              ##mutation.genes.spiked <-
              ##    sample(names(passenger.rates),length(genes.with.events.spiked),
              ##           replace=FALSE,
              ##           prob=
              ##           table(EventsBySample$MutationClass)[names(passenger.rates)])
              ##names(mutation.genes.spiked) <- genes.with.events.spiked

              genes.with.events <- c(genes.with.events, genes.with.events.spiked)
              ##mutation.genes <- c(mutation.genes, mutation.genes.spiked)

              mutation.genes <- rep("", length(genes.with.events))
              names(mutation.genes) <- genes.with.events

              for(gene in genes.with.events)
              {
                  mutation.gene <- sample(names(passenger.rates),1,replace=FALSE,
                                          prob=sim.exp.nr.mut.i[gene, ])
                  mutation.genes[gene] <- mutation.gene
                  ##add stuff to sim.Mutations.i
                  sim.Mutations.i[gene, paste("MutationsDiscovery.",mutation.gene,sep="")] <-
                      sim.Mutations.i[gene, paste("MutationsDiscovery.",
                                                  mutation.gene,sep="")] + 1

              }

              ##add stuff to sim.EventsBySample.i
              sim.EventsBySample.i <- rbind(sim.EventsBySample.i,
                                            cbind("Mut", sample,
                                                  shorten.gene.names(genes.with.events),
                                                  mutation.genes))

              mutation.genes <- rep("", length(genes.with.events))
              names(mutation.genes) <- genes.with.events
          }
      }

      ##add gene sizes and coverages to Mutations object
      sim.Mutations.i[, 51:59] <-
          sim.GeneSizes.i[rownames(sim.Mutations.i),]

      sim.Mutations.i[, 60:79] <-
          sim.Coverage.i[rownames(sim.Mutations.i),]

      ##take out rows with no mutations
      sim.Mutations.i <- sim.Mutations.i[rowSums(sim.Mutations.i[,1:25])>0,]
      sim.Mutations[[i]] <- data.frame(sim.Mutations.i)
      colnames(sim.EventsBySample.i) <- colnames(EventsBySample)
      rownames(sim.EventsBySample.i) <- NULL

      sim.EventsBySample[[i]] <- data.frame(sim.EventsBySample.i)

      ##sim.EventsBySampleBySet[[i]] <-
      ##    get.EventsBySampleBySet.from.EventsBySample(EventsBySample = sim.EventsBySample[[i]],
      ##                                                GeneSets = GeneSets,
      ##                                                GoodSamples = GoodSamples)

      sim.Scores[[i]] <- cma.scores(sim.Mutations[[i]],
                                    passenger.rates,
                                    nrow(sim.GeneSizes[[i]]))
  }

  return(list(GeneSizes = sim.GeneSizes,
              GeneSets = GeneSets,
              Coverage = sim.Coverage,
              EventsBySample = sim.EventsBySample,
              ##EventsBySampleBySet = sim.EventsBySampleBySet,
              Mutations = sim.Mutations,
              Scores = sim.Scores))
}

