##do a bunch of simulations to get p-values for the various methods
##pass.null = if TRUE, use passenger null instead of permutation null
##return.sim.data = if TRUE, return actual simulated objects as well
sim.data.p.values <- function(EventsBySample,
			      Mutations,
			      GeneSizes,
                              Coverage,
			      GeneSets,
                              passenger.rates = 
			      t(data.frame(0.55*rep(1.0e-6,25))),
			      ID2name,
			      BH = TRUE,
                              nr.iter,
			      pass.null = FALSE,
                              perc.samples = NULL,
                              spiked.set.sizes = NULL,
                              ##T/F whether to do gene method
                              gene.method = TRUE,
                              ##T/F whether to do subject method
                              perm.null.method = TRUE,
                              ##T/F whether to do heterogeneous subject method
                              perm.null.het.method = TRUE,
                              ##T/F whether to do passenger null method
                              pass.null.method = TRUE,
                              ##T/F whether to do heterogeneous passenger null method
                              pass.null.het.method = TRUE,
                              show.iter,
                              KnownMountains = c("EGFR","SMAD4","KRAS",
                              "TP53","CDKN2A","MYC","MYCN","PTEN","RB1"),
                              exclude.mountains=TRUE)
{
    ##print a random number
    ##print(paste("Random number:", runif(1)))

    ##make sure the passenger.rates object has the proper names
    names(passenger.rates) <-
        c("C.in.CpG.to.G", "C.in.CpG.to.A", "C.in.CpG.to.T",
          "G.in.CpG.to.C", "G.in.CpG.to.A", "G.in.CpG.to.T",
          "G.in.GpA.to.C", "G.in.GpA.to.A", "G.in.GpA.to.T",
          "C.in.TpC.to.G", "C.in.TpC.to.A", "C.in.TpC.to.T",
          "A.to.C", "A.to.G", "A.to.T",
          "C.not.in.CpG.or.TpC.to.G", "C.not.in.CpG.or.TpC.to.A",
          "C.not.in.CpG.or.TpC.to.T",
          "G.not.in.CpG.or.GpA.to.C", "G.not.in.CpG.or.GpA.to.A",
          "G.not.in.CpG.or.GpA.to.T",
          "T.to.C", "T.to.G", "T.to.A",
          "ins.del")
        
    ##transform the GeneSets object into a list, if it is not already a list
    GeneSets <- as.list(GeneSets)
        
    ##get the unique samples that have at least one mutation
    GoodSamples <- unique(EventsBySample[,"Sample"])

    ##get the gene symbols
    GeneNames <- rownames(GeneSizes)
    GeneSymb <- shorten.gene.names(GeneNames)

    ##if there are gene IDs which are not in the gene names, then use the ID2name to change them around
    IDs <- unique(unlist(GeneSets))
    if(length(setdiff(IDs, GeneSymb)) > 0)
    {
        GeneSets <- sapply(GeneSets,
                           function(set, i2n) {i2n[set]},
                           ID2name)
    }
    ##now eliminate all the NAs
    GeneSets <- sapply(GeneSets, function(x) {x[!is.na(x)]})

    ##EventsBySampleBySet <-
    ##    get.EventsBySampleBySet.from.EventsBySample(EventsBySample,
    ##                                                GeneSets,
    ##                                                GoodSamples)
    
    ##get simulated data
    sim.data <- sim.data(GeneSets = GeneSets,
                         GeneSizes = GeneSizes,
                         Coverage = Coverage,
                         Mutations = Mutations,
                         genes = GeneSymb,
                         passenger.rates = passenger.rates,
                         nr.iterations = nr.iter,
                         EventsBySample = EventsBySample,
                         ##EventsBySampleBySet = EventsBySampleBySet,
                         GoodSamples = GoodSamples,
                         pass.null = pass.null,
                         perc.samples = perc.samples,
                         spiked.set.sizes = spiked.set.sizes,
			 KnownMountains = KnownMountains,
			 exclude.mountains = exclude.mountains)

    all.genes <-
        shorten.gene.names(rownames(sim.data$Coverage[[1]]))

    ##get mutated spiked-in genes
    ##spiked.sets <- setdiff(names(sim.data$GeneSets),
    ##                       names(GeneSets))

    ##print("all altered genes")
    ##print(sort(rownames(sim.data$Mutations[[1]])))
    ##for(i in 1:nr.iter)
    ##{
    ##    spiked.genes.trans <- setdiff(rownames(sim.data$Coverage[[i]]),
    ##                                  rownames(Coverage))
    ##    spiked.genes.trans <-
    ##        spiked.genes.trans[grep("gene.25.25.", spiked.genes.trans)]

    ##    spiked.genes.mut <- shorten.gene.names(spiked.genes.trans)

    ##print(paste("Mutations: Iteration #", i, sep = ""))
    ##get total number of mutations and coverage of spiked-in genes from gene.set.25.25
    ##    mut.spiked.genes <-
    ##        cbind(rowSums(sim.data$Mutations[[i]][spiked.genes.trans, 1:25]),
    ##              sim.data$Mutations[[i]][spiked.genes.trans, 68])
    ##    rownames(mut.spiked.genes) <- spiked.genes.mut
    ##    colnames(mut.spiked.genes) <- c("Mutations", "Coverage")
    ##print them
    ##print(mut.spiked.genes)
    
    ##print("Samples in which these genes are altered")
    ##print(sim.data$EventsBySample[[i]][sim.data$EventsBySample[[i]]$Symbol %in%
    ##                                   spiked.genes.mut,])
    ##}

    ##get spiked-in genes
    ##spiked.genes <- setdiff(rownames(sim.data$Coverage[[1]]),
    ##                        rownames(Coverage))

    results.sim <- list()
    for(i in 1:nr.iter)
    {
        print("")
        
        if(show.iter)
        {
            print(paste("Implement methods: Iteration #", i, sep = ""))
        }

        results.sim[[i]] <-
            do.gene.set.analysis(EventsBySample =
                                 sim.data$EventsBySample[[i]],
                                 Scores =
                                 sim.data$Scores[[i]],
                                 GeneSizes =
                                 sim.data$GeneSizes[[i]],
                                 GeneSets =
                                 sim.data$GeneSets,
                                 passenger.rates = passenger.rates,
                                 Coverage =
                                 sim.data$Coverage[[i]],
                                 ID2name = ID2name,
				 BH = BH,
                                 gene.method = gene.method,
                                 perm.null.method = perm.null.method,
                                 perm.null.het.method = perm.null.het.method,
                                 pass.null.method = pass.null.method,
                                 pass.null.het.method = pass.null.het.method)
    }
    
    ##print ranks of spiked-in genes for basic and driver methods
    ##for(i in 1:nr.iter)
    ##{
    ##    print(paste("Iteration #", i, sep = ""))
    ##    ranks.basic <- rank(results.sim[[i]]$p.values.subject, ties.method = "min")
    ##    ranks.driver <- rank(results.sim[[i]]$p.values.drivers, ties.method = "min")
    ##    ranks.driver.pass <- rank(results.sim[[i]]$p.values.drivers.pass, ties.method = "min")

    ##    names(ranks.basic) <- names(ranks.driver) <- names(ranks.driver.pass) <-
    ##        rownames(results.sim[[i]])
    ##    print("ranks for basic method!")
    ##    print(ranks.basic[spiked.sets])
    ##    print("ranks for driver method!")
    ##    print(ranks.driver[spiked.sets])
    ##    print("ranks for driver method - pass!")
    ##    print(ranks.driver.pass[spiked.sets])
    ##}

    return.stuffs <- list(sim.data = sim.data,
                          results.sim = results.sim)

    return.obj <- new("SetMethodsSims")
    if(pass.null)
    {
        return.obj@null.dist <- "Passenger null"
    }
    else
    {
        return.obj@null.dist <- "Permutation null"
    }
    return.obj@perc.samples <- perc.samples
    return.obj@spiked.set.sizes <- spiked.set.sizes
    return.obj@GeneSizes <- return.stuffs$sim.data$GeneSizes
    return.obj@GeneSets <- return.stuffs$sim.data$GeneSets
    return.obj@Coverage <- return.stuffs$sim.data$Coverage
    return.obj@EventsBySample <-
        return.stuffs$sim.data$EventsBySample
    return.obj@Mutations <- return.stuffs$sim.data$Mutations
    return.obj@Scores <- return.stuffs$sim.data$Scores
    return.obj@results <- return.stuffs$results.sim

    return(return.obj)

}
